@@ -87,9 +87,16 @@ class TaskQueue:
 
     async def execute(self) -> None:
         """Execute all tasks in the queue, sending messages first and then disconnecting websockets."""
+        # It seems like there could be a race condition here.  The messages need to be sent in order,
+        # and that does seem to work.  However, if things happen really fast (i.e. a client receives
+        # a message and triggers another event before we finish sending all of these messages) then it
+        # seems possible that the messages for that second event could be intermingled with these.  I
+        # guess it's unlikely?  But I'm leaving this note here in case I ever have to debug some strange
+        # behavior with messages being received out-of-order.
         tasks = [send(websocket, message) for message, websocket in self.messages]
         if tasks:
-            await asyncio.wait(tasks)
+            for task in tasks:
+                await asyncio.wait([task])  # if we do them all at once, they can get sent out of order
         tasks = [close(websocket) for websocket in self.disconnects]
         if tasks:
             await asyncio.wait(tasks)  # TODO: not entirely sure how we handle errors that happen here
@@ -294,6 +301,17 @@ class EventHandler:
             raise ProcessingError(FailureReason.ILLEGAL_MOVE, handle=request.player.handle)
         self.handle_game_player_move_event(request.player, request.game, context.move_id)
 
+    def handle_optimal_move_request(self, request: RequestContext) -> None:
+        """Handle the Optimal Move request."""
+        log.info("Request - OPTIMAL MOVE - %s for %s", request.player.handle, request.player.game_id)
+        if not request.game:
+            raise ProcessingError(FailureReason.NOT_PLAYING, handle=request.player.handle)
+        if not request.game.is_playing():
+            raise ProcessingError(FailureReason.INVALID_GAME, comment="Game is not being played", handle=request.player.handle)
+        if not request.game.is_move_pending(request.player.handle):
+            raise ProcessingError(FailureReason.NO_MOVE_PENDING, handle=request.player.handle)
+        self.handle_game_programmatic_move_event(request.player.handle, request.game)
+
     def handle_retrieve_game_state_request(self, request: RequestContext) -> None:
         """Handle the Retrieve Game State request."""
         log.info("Request - RETRIEVE GAME - %s for %s", request.player.handle, request.player.game_id)
@@ -382,7 +400,7 @@ class EventHandler:
     def handle_player_reregistered_event(self, player: TrackedPlayer, websocket: WebSocketServerProtocol) -> None:
         """Handle the Player Registered event."""
         log.info("Event - PLAYER REREGISTERED - %s on %s", player.handle, id(websocket))
-        player.websocket = websocket
+        self.manager.retrack_player(player, websocket)
         context = PlayerRegisteredContext(handle=player.handle)
         message = Message(MessageType.PLAYER_REGISTERED, player_id=player.player_id, context=context)
         self.queue.message(message, players=[player])
@@ -394,6 +412,9 @@ class EventHandler:
         if game:
             comment = "Player %s unregistered" % player.handle
             self.handle_game_player_left_event(player, game, comment)
+        context = PlayerUnregisteredContext(handle=player.handle)
+        message = Message(MessageType.PLAYER_UNREGISTERED, context=context)
+        self.queue.message(message, players=[player])
         self.manager.delete_player(player)
 
     def handle_player_disconnected_event(self, player: TrackedPlayer) -> None:
@@ -473,9 +494,16 @@ class EventHandler:
         game.mark_active()
         player.mark_joined(game)
         game.mark_joined(player.handle)
-        context = GameJoinedContext(game_id=game.game_id)
+        context = GameJoinedContext(
+            player_handle=player.handle,
+            game_id=game.game_id,
+            name=game.name,
+            mode=game.mode,
+            advertiser_handle=game.advertiser_handle,
+        )
         message = Message(MessageType.GAME_JOINED, context=context)
         self.queue.message(message, players=[player])
+        self.handle_game_player_change_event(game, "Player joined game")
         if game.is_fully_joined():
             if self.manager.get_in_progress_game_count() >= config().in_progress_game_limit:
                 # Rather than giving the caller an error, we just ignore the game and force the
@@ -517,10 +545,10 @@ class EventHandler:
             self.queue.message(message, players=players)
             self.handle_game_state_change_event(game)
 
-    def handle_game_completed_event(self, game: TrackedGame, comment: Optional[str] = None) -> None:
+    def handle_game_completed_event(self, game: TrackedGame, winner: str, comment: str) -> None:
         """Handle the Game Completed event."""
         log.info("Event - GAME COMPLETED - %s (%s)", game.game_id, "'%s'" % comment if comment else None)
-        context = GameCompletedContext(game_id=game.game_id, comment=comment)
+        context = GameCompletedContext(game_id=game.game_id, winner=winner, comment=comment)
         message = Message(MessageType.GAME_COMPLETED, context=context)
         players = self.manager.lookup_game_players(game)
         for player in players:
@@ -596,11 +624,11 @@ class EventHandler:
         """Handle the Game Move event."""
         log.info("Event - GAME MOVE - %s for %s, move %s", handle, game.game_id, move_id)
         game.mark_active()
-        (completed, comment) = game.execute_move(handle, move_id)
+        (completed, winner, comment) = game.execute_move(handle, move_id)
+        self.handle_game_state_change_event(game)
         if completed:
-            self.handle_game_completed_event(game, comment)
+            self.handle_game_completed_event(game, winner, comment)  # type: ignore
         else:
-            self.handle_game_state_change_event(game)
             self.handle_game_next_turn_event(game)
 
     def handle_game_next_turn_event(self, game: TrackedGame) -> None:
