@@ -31,7 +31,7 @@ from attr.validators import and_, in_
 from pendulum.datetime import DateTime
 from pendulum.parser import parse
 
-from .validator import enum, length, notempty, string, stringlist
+from .validator import enum, length, notempty, regex, string, stringlist
 
 __all__ = [
     "Visibility",
@@ -53,6 +53,7 @@ __all__ = [
     "GameAction",
     "GameMove",
     "RegisterPlayerContext",
+    "ReregisterPlayerContext",
     "AdvertiseGameContext",
     "JoinGameContext",
     "ExecuteMoveContext",
@@ -61,6 +62,7 @@ __all__ = [
     "RegisteredPlayersContext",
     "AvailableGamesContext",
     "PlayerRegisteredContext",
+    "PlayerUnregisteredContext",
     "PlayerIdleContext",
     "PlayerInactiveContext",
     "PlayerMessageReceivedContext",
@@ -173,6 +175,7 @@ class MessageType(Enum):
     START_GAME = "Start Game"
     CANCEL_GAME = "Cancel Game"
     EXECUTE_MOVE = "Execute Move"
+    OPTIMAL_MOVE = "Optimal Move"
     RETRIEVE_GAME_STATE = "Retrieve Game State"
     SEND_MESSAGE = "Send Message"
 
@@ -182,6 +185,7 @@ class MessageType(Enum):
     REGISTERED_PLAYERS = "Registered Players"
     AVAILABLE_GAMES = "Available Games"
     PLAYER_REGISTERED = "Player Registered"
+    PLAYER_UNREGISTERED = "Player Unregistered"
     WEBSOCKET_IDLE = "Connection Idle"
     WEBSOCKET_INACTIVE = "Connection Inactive"
     PLAYER_IDLE = "Player Idle"
@@ -312,11 +316,12 @@ class GameStateHistory:
 
     action = attr.ib(type=str)
     color = attr.ib(type=Optional[PlayerColor])
+    card = attr.ib(type=Optional[CardType])
     timestamp = attr.ib(type=DateTime)
 
     @staticmethod
     def for_history(history: History) -> GameStateHistory:
-        return GameStateHistory(action=history.action, color=history.color, timestamp=history.timestamp)
+        return GameStateHistory(action=history.action, color=history.color, card=history.card, timestamp=history.timestamp)
 
 
 @attr.s(frozen=True)
@@ -366,19 +371,32 @@ class Context(ABC):
 MAX_HANDLE = 25
 """Maximum length of a player handle."""
 
+MAX_GAME_NAME = 40
+"""Maximum length of a game name."""
+
+HANDLE_REGEX = r"[a-zA-Z0-9_-]+"
+"""Regular expression that handles must match."""
+
 
 @attr.s(frozen=True)
 class RegisterPlayerContext(Context):
     """Context for a REGISTER_PLAYER request."""
 
-    handle = attr.ib(type=str, validator=and_(string, length(MAX_HANDLE)))
+    handle = attr.ib(type=str, validator=and_(string, length(MAX_HANDLE), regex(HANDLE_REGEX)))
+
+
+@attr.s(frozen=True)
+class ReregisterPlayerContext(Context):
+    """Context for a REREGISTER_PLAYER request."""
+
+    handle = attr.ib(type=str, validator=and_(string, length(MAX_HANDLE), regex(HANDLE_REGEX)))
 
 
 @attr.s(frozen=True)
 class AdvertiseGameContext(Context):
     """Context for an ADVERTISE_GAME request."""
 
-    name = attr.ib(type=str, validator=string)
+    name = attr.ib(type=str, validator=and_(string, length(MAX_GAME_NAME)))
     mode = attr.ib(type=GameMode, validator=enum(GameMode))
     players = attr.ib(type=int, validator=in_([2, 3, 4]))
     visibility = attr.ib(type=Visibility, validator=enum(Visibility))
@@ -437,6 +455,13 @@ class PlayerRegisteredContext(Context):
     handle = attr.ib(type=str)
 
 
+@attr.s(frozen=True)
+class PlayerUnregisteredContext(Context):
+    """Context for a PLAYER_UNREGISTERED event."""
+
+    handle = attr.ib(type=str)
+
+
 @attr.s(frozen=True)
 class PlayerIdleContext(Context):
     """Context for a PLAYER_IDLE event."""
@@ -478,8 +503,11 @@ class GameInvitationContext(Context):
 class GameJoinedContext(Context):
     """Context for a GAME_JOINED event."""
 
+    player_handle = attr.ib(type=str)
     game_id = attr.ib(type=str)
-    # TODO: we need the handle of the player that joined, in case multiple players share the same websocket
+    name = attr.ib(type=str)
+    mode = attr.ib(type=GameMode)
+    advertiser_handle = attr.ib(type=str)
 
 
 @attr.s(frozen=True)
@@ -503,7 +531,8 @@ class GameCompletedContext(Context):
     """Context for a GAME_COMPLETED event."""
 
     game_id = attr.ib(type=str)
-    comment = attr.ib(type=Optional[str])
+    winner = attr.ib(type=str)
+    comment = attr.ib(type=str)
 
 
 @attr.s(frozen=True)
@@ -552,7 +581,7 @@ class GameStateChangeContext(Context):
         player = GameStatePlayer.for_player(view.player)
         recent_history = [GameStateHistory.for_history(entry) for entry in history]
         opponents = [GameStatePlayer.for_player(opponent) for opponent in view.opponents.values()]
-        return GameStateChangeContext(game_id, recent_history, player, opponents)
+        return GameStateChangeContext(game_id=game_id, recent_history=recent_history, player=player, opponents=opponents,)
 
 
 @attr.s(frozen=True)
@@ -586,6 +615,7 @@ _PLAYER_ID: Dict[MessageType, bool] = {
     MessageType.START_GAME: True,
     MessageType.CANCEL_GAME: True,
     MessageType.EXECUTE_MOVE: True,
+    MessageType.OPTIMAL_MOVE: True,
     MessageType.RETRIEVE_GAME_STATE: True,
     MessageType.SEND_MESSAGE: True,
     MessageType.SERVER_SHUTDOWN: False,
@@ -595,6 +625,7 @@ _PLAYER_ID: Dict[MessageType, bool] = {
     MessageType.REGISTERED_PLAYERS: False,
     MessageType.AVAILABLE_GAMES: False,
     MessageType.PLAYER_REGISTERED: True,
+    MessageType.PLAYER_UNREGISTERED: False,
     MessageType.PLAYER_IDLE: False,
     MessageType.PLAYER_INACTIVE: False,
     MessageType.PLAYER_MESSAGE_RECEIVED: False,
@@ -615,7 +646,7 @@ _PLAYER_ID: Dict[MessageType, bool] = {
 # Map from MessageType to context
 _CONTEXT: Dict[MessageType, Optional[Type[Context]]] = {
     MessageType.REGISTER_PLAYER: RegisterPlayerContext,
-    MessageType.REREGISTER_PLAYER: None,
+    MessageType.REREGISTER_PLAYER: ReregisterPlayerContext,
     MessageType.UNREGISTER_PLAYER: None,
     MessageType.LIST_PLAYERS: None,
     MessageType.ADVERTISE_GAME: AdvertiseGameContext,
@@ -625,6 +656,7 @@ _CONTEXT: Dict[MessageType, Optional[Type[Context]]] = {
     MessageType.START_GAME: None,
     MessageType.CANCEL_GAME: None,
     MessageType.EXECUTE_MOVE: ExecuteMoveContext,
+    MessageType.OPTIMAL_MOVE: None,
     MessageType.RETRIEVE_GAME_STATE: None,
     MessageType.SEND_MESSAGE: SendMessageContext,
     MessageType.SERVER_SHUTDOWN: None,
@@ -634,6 +666,7 @@ _CONTEXT: Dict[MessageType, Optional[Type[Context]]] = {
     MessageType.REGISTERED_PLAYERS: RegisteredPlayersContext,
     MessageType.AVAILABLE_GAMES: AvailableGamesContext,
     MessageType.PLAYER_REGISTERED: PlayerRegisteredContext,
+    MessageType.PLAYER_UNREGISTERED: PlayerUnregisteredContext,
     MessageType.PLAYER_IDLE: PlayerIdleContext,
     MessageType.PLAYER_INACTIVE: PlayerInactiveContext,
     MessageType.PLAYER_MESSAGE_RECEIVED: PlayerMessageReceivedContext,
