@@ -35,7 +35,7 @@ from __future__ import annotations  # see: https://stackoverflow.com/a/33533514/
 import asyncio
 import logging
 import random
-from typing import Callable, Dict, List, Optional, Tuple
+from typing import Dict, List, Optional, Tuple
 from uuid import uuid4
 
 import attr
@@ -43,7 +43,7 @@ import pendulum
 from apologies.engine import Character, Engine
 from apologies.game import GameMode, History, PlayerColor, PlayerView
 from apologies.rules import Move
-from apologies.source import CharacterInputSource
+from apologies.source import NoOpInputSource
 from ordered_set import OrderedSet  # this makes expected results easier to articulate in test code
 from pendulum.datetime import DateTime
 from websockets import WebSocketServerProtocol
@@ -194,22 +194,6 @@ class TrackedPlayer:
         self.connection_state = ConnectionState.DISCONNECTED
 
 
-@attr.s
-class NoOpInputSource(CharacterInputSource):
-    """A no-op input source, which raises an error if ever used."""
-
-    # The Apologies library is designed with a synchronous callback model in mind.
-    # Since this server uses Python's asyncio instead, we ignore the callback and
-    # trigger individual steps in the game play process when events are received.
-    # This implementation throws an error if it's ever used, to make it clear if
-    # we do something wrong.
-
-    def choose_move(
-        self, _mode: GameMode, _view: PlayerView, _moves: List[Move], _evaluator: Callable[[PlayerView, Move], PlayerView]
-    ) -> Move:
-        raise NotImplementedError
-
-
 @attr.s(frozen=True)
 class CurrentTurn:
 
@@ -301,8 +285,8 @@ class TrackedEngine:
             raise ProcessingError(FailureReason.INTERNAL_ERROR, "Illegal state for operation")
         return handle == self._current.handle and move_id in self._current.movedict
 
-    def execute_move(self, handle: str, move_id: str) -> Tuple[bool, Optional[str]]:
-        """Execute a player's move, returning whether the game was completed (and a comment if so)."""
+    def execute_move(self, handle: str, move_id: str) -> Tuple[bool, Optional[str], Optional[str]]:
+        """Execute a player's move, returning whether the game was completed (and the winner and a comment if so)."""
         if (
             not self._engine
             or not self._current
@@ -317,10 +301,10 @@ class TrackedEngine:
         if self._engine.completed:
             self._current = None
             character, player = self._engine.winner()  # type: ignore
-            comment = "Player %s (%s) won after %d turns" % (character.name, player.color.name, player.turns)
-            return True, comment
+            comment = "Player %s won after %d turns" % (character.name, player.turns)
+            return True, character.name, comment
         self._current = CurrentTurn.next_player(self._engine) if done else self._current.draw_again(self._engine)
-        return False, None
+        return False, None, None
 
 
 # pylint: disable=too-many-instance-attributes,too-many-public-methods
@@ -375,6 +359,11 @@ class TrackedGame:
             invited_handles=context.invited_handles[:],
         )
 
+    @property
+    def completed(self) -> bool:
+        """Whether the game is completed."""
+        return self.completed_date is not None  # this way, we don't have to mess with which statuses mean "completed"
+
     def to_advertised_game(self) -> AdvertisedGame:
         """Convert this tracked game to an AdvertisedGame."""
         return AdvertisedGame(
@@ -526,7 +515,7 @@ class TrackedGame:
         else:
             raise ProcessingError(FailureReason.INTERNAL_ERROR, "Illegal state for operation")
 
-    def execute_move(self, handle: str, move_id: str) -> Tuple[bool, Optional[str]]:
+    def execute_move(self, handle: str, move_id: str) -> Tuple[bool, Optional[str], Optional[str]]:
         """Execute a player's move, returning an indication of whether the game was completed (and a comment if so)."""
         return self._engine.execute_move(handle, move_id)
 
@@ -580,10 +569,13 @@ class StateManager:
 
     def mark_active(self, player: TrackedPlayer) -> None:
         """Mark a player and its associated websocket as active."""
-        if not player.websocket or player.websocket not in self._websocket_map:
-            raise ProcessingError(FailureReason.INTERNAL_ERROR, comment="Did not find player websocket", handle=player.handle)
-        websocket = self._websocket_map[player.websocket]
-        websocket.mark_active()
+        if player.websocket:
+            # The player might not have a websocket if they've been disconnected.
+            # However, if the player has a websocket, then it must be valid or we have a problem of internal consistency.
+            if player.websocket not in self._websocket_map:
+                raise ProcessingError(FailureReason.INTERNAL_ERROR, comment="Did not find player websocket", handle=player.handle)
+            websocket = self._websocket_map[player.websocket]
+            websocket.mark_active()
         player.mark_active()
 
     def track_websocket(self, websocket: WebSocketServerProtocol) -> None:
@@ -664,11 +656,18 @@ class StateManager:
         if handle in self._handle_map:
             raise ProcessingError(FailureReason.DUPLICATE_USER, handle=handle)
         player_id = "%s" % uuid4()
+        self._websocket_map[websocket].mark_active()  # we never want a websocket to time out before an associated player
         self._websocket_map[websocket].player_ids.add(player_id)
         self._player_map[player_id] = TrackedPlayer.for_context(player_id, websocket, handle)
         self._handle_map[handle] = player_id
         return self._player_map[player_id]
 
+    def retrack_player(self, player: TrackedPlayer, websocket: WebSocketServerProtocol) -> None:
+        """Re-track and existing player, associating it with a different websocket."""
+        player.websocket = websocket
+        self._websocket_map[websocket].mark_active()  # we never want a websocket to time out before an associated player
+        self._websocket_map[websocket].player_ids.add(player.player_id)
+
     def delete_player(self, player: TrackedPlayer) -> None:
         """Delete a tracked player, so it is no longer tracked."""
         if player.websocket and player.websocket in self._websocket_map:
@@ -714,14 +713,15 @@ class StateManager:
         """Look up the last active date for all games."""
         result: List[Tuple[TrackedGame, DateTime]] = []
         for game in self._game_map.values():
-            result.append((game, game.last_active_date))
+            if not game.completed:
+                result.append((game, game.last_active_date))
         return result
 
     def lookup_game_completion(self) -> List[Tuple[TrackedGame, Optional[DateTime]]]:
         """Look up the completed date for all completed games."""
         result: List[Tuple[TrackedGame, Optional[DateTime]]] = []
         for game in self._game_map.values():
-            if game.game_state == GameState.COMPLETED:
+            if game.completed:
                 result.append((game, game.completed_date))
         return result
 
